#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#ifndef X11
#include <windows.h>
#else
#include <GL/glx.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/extensions/XShm.h>
#include <Xm/MwmUtil.h>
#include <sys/time.h>
#endif
#include <GL/gl.h>
#include "plane.h"
#ifndef OGL1
#include "texture.h"
#endif
#include "glext.h"

#ifdef X11
#define DWORD unsigned long
#endif

#define PI 3.141593f

#define SC_ROTATING_TRANGLE 1
#define SC_TRIANGLES_DRAWCALLS 2
#define SC_TRIANGLE_PIXEL_RATE 3

FILE *logFile, *csvFile;

int windowWidth = 640;
int windowHeight = 480;
int colorsBits = 24;
int depthBits = 16;                        // 0 turns off z-buffer
int windowFrame = 1;
/*int timeDemo = 0;
int targetFps = 60;*/
int clearBuffers = 1;
int scene = 1;
int trianglesCount = 100;
int scene2_trianglesPixels = 10000;
int scene3_planesCount = 4;
int wireFrame = 0;                        // 1 = On, 2 = On Anti-Aliased; -1 and -2 for overlay wireframe
int blending = 0;
char csvFileName[256];
int csvHeaders;
int runSeconds;
int triangleStrip = 1;
char textureFileName1[256];
char textureFileName2[256];
char testName[256];
int polygonNoColor = 0;
int textureFilter = 1;                    // 0 = nearest, 1 = bilinear
int disableMultitexture = 0;
int depthTestLogging = 0;
int checkDephtTestFill = 0;
int checkDephtTestWireframe = 0;

#ifndef X11
HWND hWnd;
HDC hDC;
int iFormat;
HGLRC hRC;        
#else
Display* dpy;
XVisualInfo* vi;
XSetWindowAttributes swa;
Window win;
GLXContext cx;
#endif

#ifndef X11
void Init(HINSTANCE hInstance, LPSTR lpCmdLine);
#else
void Init(char* lpCmdLine);
#endif
void Run();
void Draw(int scene, float param);
void DeInit();

#ifndef X11
LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
#else
int process_input(Display *dpy);
#endif

int sceneInfoLogged = 0;
float fps = -1.0f;
GLuint *displayLists = 0;
int displayListsCount = 0;
float circleRadius = 0.0f;
GLuint texture1, texture2 = 0;
int realTrianglesCount = 0;
int isMultitexture = 0;
#ifndef X11
PROC glActiveTextureARB = 0;
PROC glMultiTexCoord2fARB = 0;
#endif
int supportedBgr = 0;
const char* programInfo = "GPUbench 1.1 (c) Ladislav Zima, Jiri Zima";

#ifndef X11
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int iCmdShow) {
    Init(hInstance, lpCmdLine);
#else
int main(int argc, char** argv) {
    char* lpCmdLine = argc >= 2 ? argv[1] : "";
    Init(lpCmdLine);
#endif
    Run();
    DeInit();
}

#ifndef X11
void Init(HINSTANCE hInstance, LPSTR lpCmdLine) {
#else
void Init(char* lpCmdLine) {
#endif
    FILE *configFile;
    char word[64];
    char fullProgramInfo[256];
    const char *str;
        
    logFile = fopen("gpubench.log", "w");
    fprintf(logFile, "cmdLine=%s\n\n", lpCmdLine);

    csvFileName[0] = 0; textureFileName1[0] = 0; textureFileName2[0] = 0; testName[0] = 0;
    configFile = fopen(lpCmdLine && strcmp(lpCmdLine, "") != 0 ? lpCmdLine : "gpubench.cfg", "r");
    while (fscanf(configFile, "%s", word) != EOF) {
        //fprintf(logFile, "word=%s\n", word);
        
        if (strcmp(word, "windowWidth") == 0)
            fscanf(configFile, "%i", &windowWidth);
        else if (strcmp(word, "windowHeight") == 0)
            fscanf(configFile, "%i", &windowHeight);
        else if (strcmp(word, "colorsBits") == 0)
            fscanf(configFile, "%i", &colorsBits);
        else if (strcmp(word, "depthBits") == 0)
            fscanf(configFile, "%i", &depthBits);
        else if (strcmp(word, "windowFrame") == 0)
            fscanf(configFile, "%i", &windowFrame);
        /*else if (strcmp(word, "timeDemo") == 0)
            fscanf(configFile, "%i", &timeDemo);
        else if (strcmp(word, "targetFps") == 0)
            fscanf(configFile, "%i", &targetFps);*/
        else if (strcmp(word, "clearBuffers") == 0)
            fscanf(configFile, "%i", &clearBuffers);
        else if (strcmp(word, "scene") == 0)
            fscanf(configFile, "%i", &scene);
        else if (strcmp(word, "trianglesCount") == 0)
            fscanf(configFile, "%i", &trianglesCount);
        else if (strcmp(word, "scene2_trianglesPixels") == 0)
            fscanf(configFile, "%i", &scene2_trianglesPixels);
        else if (strcmp(word, "scene3_planesCount") == 0)
            fscanf(configFile, "%i", &scene3_planesCount);
        else if (strcmp(word, "wireFrame") == 0)
            fscanf(configFile, "%i", &wireFrame);
        else if (strcmp(word, "blending") == 0)
            fscanf(configFile, "%i", &blending);
        else if (strcmp(word, "csvFileName") == 0)
            fscanf(configFile, "%s", csvFileName);
        else if (strcmp(word, "csvHeaders") == 0)
            fscanf(configFile, "%i", &csvHeaders);
        else if (strcmp(word, "runSeconds") == 0)
            fscanf(configFile, "%i", &runSeconds);
        else if (strcmp(word, "triangleStrip") == 0)
            fscanf(configFile, "%i", &triangleStrip);
        else if (strcmp(word, "textureFileName1") == 0)
            fscanf(configFile, "%s", textureFileName1);
        else if (strcmp(word, "textureFileName2") == 0)
            fscanf(configFile, "%s", textureFileName2);
        else if (strcmp(word, "testName") == 0)
            fscanf(configFile, "%s", testName);
        else if (strcmp(word, "polygonNoColor") == 0)
            fscanf(configFile, "%i", &polygonNoColor);
        else if (strcmp(word, "textureFilter") == 0)
            fscanf(configFile, "%i", &textureFilter);
        else if (strcmp(word, "disableMultitexture") == 0)
            fscanf(configFile, "%i", &disableMultitexture);
        else if (strcmp(word, "depthTestLogging") == 0)
            fscanf(configFile, "%i", &depthTestLogging);
        else if (strcmp(word, "checkDephtTestFill") == 0)
            fscanf(configFile, "%i", &checkDephtTestFill);
    	else if (strcmp(word, "checkDephtTestWireframe") == 0)
            fscanf(configFile, "%i", &checkDephtTestWireframe);
    }
    fclose(configFile);

    sprintf(fullProgramInfo, "%s 2018", programInfo);

#ifndef X11
    WNDCLASS wc;

    wc.style = CS_OWNDC;
    wc.lpfnWndProc = WndProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hInstance = hInstance;
    wc.hIcon = LoadIcon(NULL, IDI_APPLICATION);
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
    wc.lpszMenuName = NULL;
    wc.lpszClassName = "GPUbench";
    RegisterClass(&wc);

    int desktopWidth = GetSystemMetrics(SM_CXSCREEN);
    int desktopHeight = GetSystemMetrics(SM_CYSCREEN);
    
    int windowStyle = windowFrame ? WS_CAPTION | WS_POPUPWINDOW | WS_VISIBLE | WS_MINIMIZEBOX : WS_POPUP | WS_VISIBLE | WS_SYSMENU;

    RECT wr = {0, 0, windowWidth, windowHeight};                // compute the window size for correct client area
    AdjustWindowRect(&wr, windowStyle, FALSE);
    
    int windowLeft = windowFrame ? (desktopWidth - (wr.right - wr.left)) / 2 : 0;
    int windowTop = windowFrame ? (desktopHeight - (wr.bottom - wr.top)) / 2 : 0;
    hWnd = CreateWindow("GPUbench", fullProgramInfo, windowStyle, windowLeft, windowTop, wr.right - wr.left, wr.bottom - wr.top, NULL, NULL, hInstance, NULL);

    PIXELFORMATDESCRIPTOR pfd;
    
    hDC = GetDC(hWnd);

    ZeroMemory(&pfd, sizeof(pfd));
    pfd.nSize = sizeof(pfd);
    pfd.nVersion = 1;
    pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
    pfd.iPixelType = PFD_TYPE_RGBA;
    pfd.cColorBits = colorsBits;
    pfd.cDepthBits = depthBits;
    pfd.iLayerType = PFD_MAIN_PLANE;
    iFormat = ChoosePixelFormat(hDC, &pfd);
    SetPixelFormat (hDC, iFormat, &pfd);

    hRC = wglCreateContext(hDC);
    wglMakeCurrent(hDC, hRC);

    glViewport(0, 0, windowWidth, windowHeight);
#else
    {
    Screen* scr;
    int attributeList[] = { GLX_RGBA, 
        GLX_RED_SIZE, 1, /*colorsBits == 8 ? 3 : (colorsBits == 16 ? 5 : 8), 
        GLX_GREEN_SIZE, colorsBits == 8 ? 3 : (colorsBits == 16 ? 5 : 8), 
        GLX_BLUE_SIZE, colorsBits == 8 ? 2 : (colorsBits == 16 ? 5 : 8), 
        GLX_ALPHA_SIZE, colorsBits == 8 ? 0 : (colorsBits == 16 ? 0 : 0),*/
        GLX_DEPTH_SIZE, 0 /*placeholder for depthBits*/, 
        GLX_DOUBLEBUFFER, None };
       // window positioning and sizing from Quake 2 :-)
    MotifWmHints wmhints;
    Atom aHints;
    XSizeHints sizehints;
    XWindowChanges changes;
    attributeList[4]=depthBits;

    dpy = XOpenDisplay(0);
    scr = ScreenOfDisplay(dpy, 0);


    //printf("GLX_RED_SIZE=%d, GLX_GREEN_SIZE=%d, GLX_BLUE_SIZE=%d, GLX_ALPHA_SIZE=%d, GLX_DEPTH_SIZE=%d\n", 
    //    colorsBits == 8 ? 3 : (colorsBits == 16 ? 5 : 8), colorsBits == 8 ? 3 : (colorsBits == 16 ? 5 : 8),
    //    colorsBits == 8 ? 2 : (colorsBits == 16 ? 5 : 8), colorsBits == 8 ? 0 : (colorsBits == 16 ? 0 : 0),
    //    depthBits);
    // SGI O2 (1996) can do RGBA8888 and RGBA5550; can't do RGBA5650 and RGBA5551
    if (!(vi = glXChooseVisual(dpy, DefaultScreen(dpy), attributeList))) {
        fprintf(stderr, "No suitable RGB visual available.\n");
        exit(EXIT_FAILURE);
    }

    cx = glXCreateContext(dpy, vi, 0, GL_TRUE);
    //BOOL isdirect = glXIsDirect(dpy, cx);

    swa.colormap = XCreateColormap(dpy, RootWindow(dpy, vi->screen), vi->visual, AllocNone);

    swa.border_pixel = 0;
    swa.event_mask = ExposureMask | StructureNotifyMask | KeyPressMask;
    win = XCreateWindow(dpy, RootWindow(dpy, vi->screen), 
        windowFrame == 1 ? (scr->width - windowWidth) / 2 : 0, windowFrame == 1 ? (scr->height - windowHeight) / 2 : 0, 
        windowWidth, windowHeight, 0, vi->depth, InputOutput, vi->visual, CWBorderPixel | CWColormap | CWEventMask, &swa);
        

    aHints = XInternAtom(dpy, "_MOTIF_WM_HINTS", 0);
    if (aHints == None)
    {
        printf("Could not intern X atom for _MOTIF_WM_HINTS.\n");
    } else {
        if (windowFrame == 0) {
            wmhints.flags = MWM_HINTS_DECORATIONS;
            wmhints.decorations = 0;     // Absolutely no decorations.
        }
        XChangeProperty(dpy, win, aHints, aHints, 32, PropModeReplace, (unsigned char *)&wmhints, 4);

        sizehints.flags = USPosition | (windowFrame == 0 ? USSize : 0);
        sizehints.x = windowFrame == 1 ? (scr->width - windowWidth) / 2 : 0;
        sizehints.y = windowFrame == 1 ? (scr->height - windowHeight) / 2 : 0;
        if (windowFrame == 0) {
            sizehints.width = windowWidth;
            sizehints.height = windowHeight;
        }
        XSetWMNormalHints(dpy, win, &sizehints);

        changes.x = windowFrame == 1 ? (scr->width - windowWidth) / 2 : 0;
        changes.y = windowFrame == 1 ? (scr->height - windowHeight) / 2 : 0;
        if (windowFrame == 0) {
            changes.width = windowWidth;
            changes.height = windowHeight;
        }
        changes.stack_mode = TopIf;
        XConfigureWindow(dpy, win,
             CWX | CWY | (windowFrame == 0 ? CWWidth | CWHeight : 0) | CWStackMode,
             &changes);
    }
        
    XStoreName(dpy, win, fullProgramInfo);
    XMapWindow(dpy, win);

    glXMakeCurrent(dpy, win, cx);

    XSynchronize(dpy, 0);
    //glXSwapIntervalSGI(0);
    }
#endif
    
    fprintf(logFile, "windowWidth=%i, windowHeight=%i, colorsBits=%i, depthBits=%i, windowFrame=%i, " /*"timeDemo=%i, targetFps=%i, "*/ "clearBuffers=%i, scene=%i, trianglesCount=%i, "
        "scene2_trianglesPixels=%i, scene3_planesCount=%i, wireFrame=%i, blending=%i, csvFileName=%s, csvHeaders=%i, runSeconds=%i, triangleStrip=%i, textureFileName1=%s, textureFileName2=%s, "
        "testName=%s, polygonNoColor=%i, textureFilter=%i, disableMultitexture=%i, depthTestLogging=%i, checkDephtTestFill=%i, checkDephtTestWireframe=%i\n\n", 
        windowWidth, windowHeight, colorsBits, depthBits, windowFrame, /*timeDemo, targetFps,*/ clearBuffers, scene, trianglesCount, 
        scene2_trianglesPixels, scene3_planesCount, wireFrame, blending, csvFileName, csvHeaders, runSeconds, triangleStrip, textureFileName1, textureFileName2,
        testName, polygonNoColor, textureFilter, disableMultitexture, depthTestLogging, checkDephtTestFill, checkDephtTestWireframe);
        
    str = glGetString(GL_VENDOR);
    fprintf(logFile, "GL_VENDOR: %s\n", str);
    str = glGetString(GL_RENDERER);
    fprintf(logFile, "GL_RENDERER: %s\n", str);
#ifndef X11
	PIXELFORMATDESCRIPTOR pfd2;
	DescribePixelFormat(hDC, iFormat, sizeof(pfd2), &pfd2);
	int generic_format = pfd2.dwFlags & PFD_GENERIC_FORMAT;
	int generic_accelerated = pfd2.dwFlags & PFD_GENERIC_ACCELERATED;
	char gldriver[256];
	gldriver[0] = 0;
	if (generic_format && !generic_accelerated)
		sprintf(gldriver, "%s", "Software");
	else if (generic_format && generic_accelerated)
		sprintf(gldriver, "%s", "Hardware - MCD");
	else if (! generic_format && ! generic_accelerated)
		sprintf(gldriver, "%s", "Hardware - ICD");
    fprintf(logFile, "OpenGL Driver Model: %s\n", gldriver);
#endif
    str = glGetString(GL_VERSION);
    fprintf(logFile, "GL_VERSION: %s\n\n", str);    
    str = glGetString(GL_EXTENSIONS);
    fprintf(logFile, "GL_EXTENSIONS:\n%s\n\n", str);    

    isMultitexture = strstr(str, "GL_ARB_multitexture") != 0 && !disableMultitexture;
#ifndef X11
    if (isMultitexture) {
        glActiveTextureARB = (PROC)wglGetProcAddress("glActiveTextureARB");
        glMultiTexCoord2fARB = (PROC)wglGetProcAddress("glMultiTexCoord2fARB");
    }
#endif
    
    supportedBgr = strstr(str, "GL_EXT_bgra") != 0;

    if (csvFileName[0] != 0) {
        csvFile = fopen(csvFileName, "a");

        if (csvHeaders) {
            fprintf(csvFile, "sep=,\ntestName,fps,fillrate,trianglerate,windowWidth,windowHeight,colorsBits,depthBits,windowFrame," /*"timeDemo,targetFps,"*/ "clearBuffers,scene,trianglesCount,realTrianglesCount,"
                "scene2_trianglesPixels,scene3_planesCount,wireFrame,blending,runSeconds,triangleStrip,textureFileName1,textureFileName2,polygonNoColor,textureFilter,isMultitexture,"
                "disableMultitexture\n");
        }

        realTrianglesCount = 
            scene == SC_TRIANGLES_DRAWCALLS ? ((int)sqrt(trianglesCount) / 2 * 2) * ((int)sqrt(trianglesCount) / 2 * 2) : 
            (scene == SC_TRIANGLE_PIXEL_RATE ? (int)sqrt(trianglesCount / 2) * (int)sqrt(trianglesCount / 2) * 2 : 
            1);

        fprintf(csvFile, "%s"/*",%i,%i,%i,%i,%i," /_*"%i,%i,"*_/ "%i,%i,%i,%i,"
            "%i,%i,%i,%i,%i,%i,%s,%s,%i,%i,%i,"
            "%i"*/, 
            testName/*, windowWidth, windowHeight, colorsBits, depthBits, windowFrame, /_*timeDemo, targetFps,*_/ clearBuffers, scene, trianglesCount, realTrianglesCount,
            scene2_trianglesPixels, scene3_planesCount, wireFrame, blending, runSeconds, triangleStrip, textureFileName1, textureFileName2, polygonNoColor, textureFilter, isMultitexture,
            disableMultitexture*/);
    }
}

#ifdef X11
DWORD GetTickCount() {
    struct timeval te;
    gettimeofday(&te, NULL);
    return (DWORD)(te.tv_sec * 1000LL + te.tv_usec / 1000);
}
#endif

void Run() {
#ifndef X11
    MSG msg;
#endif

    DWORD lastTicks = GetTickCount(), anchor=lastTicks;
    float delay = 0.0f;
    //float fixedDelay = 1.0f / targetFps;
    
    float longDelay = 0.0f, totalDelay = 0.0f;
    int frames = 0;

    float param = 0.0f;
    float direction = 1.0f;

    int afterFirstFrame = 0;
    while (1) {
#ifndef X11
        if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
            if (msg.message == WM_QUIT) {
                break;
            } else {
                TranslateMessage(&msg);
                DispatchMessage(&msg);
            }
        } else {
#else
        if((int)(lastTicks-anchor)>300){
          if(!process_input(dpy)) break;
          anchor=lastTicks;
        }
        XSync(dpy,0);
        {
#endif
            DWORD ticks = GetTickCount();
            delay = (ticks - lastTicks) * .001f;
            lastTicks = ticks;
            
            totalDelay += delay;
            if (runSeconds > 0 && totalDelay + 0.1f > (float)runSeconds && afterFirstFrame)
                break;
            
            longDelay += delay;
            frames++;
            if (longDelay > 1.0f) {
                fps = frames / longDelay;
                longDelay = 0.0f;
                frames = 0;
            }
            
            if (fps != -1.0f) {
                char windowTitle[512];
                sprintf(windowTitle, "%s - %.2f fps", programInfo, fps);
#ifndef X11
                SetWindowText(hWnd, windowTitle);
#else
                XStoreName(dpy, win, windowTitle);
#endif
            }
            
            if (scene == SC_ROTATING_TRANGLE) {
                param += 180.0f * delay;
            }
            else if (scene == SC_TRIANGLES_DRAWCALLS) {
                param += (scene == 2 ? 45.0f : 180.0f) * delay * direction;
                if (param > 22.5f || param < -22.5f) {
                    param = 22.5f * direction;
                    direction *= -1.0f;
                }
            }
            else if (scene == SC_TRIANGLE_PIXEL_RATE) {
                param -= 2.0f * PI * delay;
            }
            
            Draw(scene, param);
            
#ifndef X11
            SwapBuffers(hDC);
#else
            glXSwapBuffers(dpy, win);
#endif
            
            if (!afterFirstFrame) {
                afterFirstFrame = 1;

                lastTicks = GetTickCount();
                delay = 0.0f;
                //float fixedDelay = 1.0f / targetFps;
                
                longDelay = 0.0f, totalDelay = 0.0f;
                frames = 0;
            }
        }
    }
}

void Draw(int scene, float param) {
    glClearColor (0.0f, 0.0f, 0.0f, 0.0f);
//    if (clearBuffers)
//        glClear(GL_COLOR_BUFFER_BIT | (depthBits ? GL_DEPTH_BUFFER_BIT : 0));
    if (depthBits == 0) {
        if (clearBuffers) glClear(GL_COLOR_BUFFER_BIT);
        glDisable(GL_DEPTH_TEST); if (depthTestLogging) fprintf(logFile, "    glDisable(GL_DEPTH_TEST);\n");
        glDepthFunc(GL_ALWAYS); if (depthTestLogging) fprintf(logFile, "        glDepthFunc(GL_ALWAYS);\n");
    } else {
    glClear((clearBuffers ? GL_COLOR_BUFFER_BIT : 0) | GL_DEPTH_BUFFER_BIT);
		glEnable(GL_DEPTH_TEST); if (depthTestLogging) fprintf(logFile, "    glEnable(GL_DEPTH_TEST);\n");
		glDepthFunc(GL_LEQUAL); if (depthTestLogging) fprintf(logFile, "        glDepthFunc(GL_LEQUAL);\n");
	}
        
    if (wireFrame) {
          glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
          
          if (wireFrame == 2) {
            glEnable(GL_LINE_SMOOTH);
            //glEnable(GL_POLYGON_SMOOTH);
            glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
            //glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);

            glEnable(GL_BLEND);
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        }
    }

    if (scene == SC_ROTATING_TRANGLE) {
        glPushMatrix();
        glRotatef(param, 0.0f, 0.0f, 1.0f);
        glBegin(GL_TRIANGLES);
            glColor3f(1.0f, 0.0f, 0.0f);    glVertex2f(0.0f, 1.0f);
            glColor3f(0.0f, 1.0f, 0.0f);    glVertex2f(0.87f, -0.5f);
            glColor3f(0.0f, 0.0f, 1.0f);    glVertex2f(-0.87f, -0.5f);
        glEnd();
        glPopMatrix();
    }
    else if (scene == SC_TRIANGLES_DRAWCALLS) {
        int x, y, trianglesCountSR, trianglesCountSRHalf;
        float triangleSide,trianglePositionStep;
        if (depthBits) {
            glDepthFunc(GL_LEQUAL); if (depthTestLogging) fprintf(logFile, "        glDepthFunc(GL_LEQUAL);\n");
        }

        glMatrixMode(GL_PROJECTION);
        glPushMatrix();
        glOrtho(-windowWidth * 0.5, windowWidth * 0.5, -windowHeight * 0.5, windowHeight * 0.5, -1.0, 1.0);
        glMatrixMode(GL_MODELVIEW);

        trianglesCountSR = sqrt(trianglesCount);
        trianglesCountSRHalf = trianglesCountSR / 2;            // CAUTION: the amount of triangles per side is rounded to 2
        triangleSide = sqrtf(scene2_trianglesPixels * 2.0f);
        trianglePositionStep = 1.0f / (trianglesCountSR + 2);
        
        if (!sceneInfoLogged) {
            fprintf(logFile, "trianglesPerSide=%i, trianglesTotal=%i\n", trianglesCountSRHalf * 2, (trianglesCountSRHalf * 2) * (trianglesCountSRHalf * 2));
            sceneInfoLogged = 1;
        }
        
        for (y = trianglesCountSRHalf; y > -trianglesCountSRHalf; y--) {
            for (x = trianglesCountSRHalf; x > -trianglesCountSRHalf; x--) {
                glPushMatrix();
                glTranslatef(((float)x - 0.5f) * windowWidth * trianglePositionStep, ((float)y - 0.5f) * windowHeight * trianglePositionStep, 0.0f);
                glRotatef(param, 0.0f, 0.0f, 1.0f);
                glScalef(triangleSide, triangleSide, 0.0f);
                glBegin(GL_TRIANGLES);
                {
                    int c = ((-1) * (x - trianglesCountSRHalf) + (y + trianglesCountSRHalf)) % 6;    // R, G, B, RG, GB, RB
                    float r = c == 0 || c == 3 || c == 5 ? 1.0f : 0.0f;
                    float g = c == 1 || c == 3 || c == 4 ? 1.0f : 0.0f;
                    float b = c == 2 || c == 4 || c == 5 ? 1.0f : 0.0f;
                    
                    glColor3f(r * 0.5f, g * 0.5f, b * 0.5f);    glVertex2f(-0.5f, -0.5f);
                    glColor3f(r * 0.0f, g * 0.0f, b * 0.0f);    glVertex2f(0.5f, -0.5f);
                    glColor3f(r * 1.0f, g * 1.0f, b * 1.0f);    glVertex2f(0.0f, 0.5f);
                }
                glEnd();
                glPopMatrix();
            }
        }

        glMatrixMode(GL_PROJECTION);
        glPopMatrix();
        glMatrixMode(GL_MODELVIEW);
    }
    else if (scene == SC_TRIANGLE_PIXEL_RATE) {
#ifndef OGL1
        if (textureFileName1[0] != 0 && texture1 == 0) {
            glEnable(GL_TEXTURE_2D); 

            texture1 = LoadTexture(textureFileName1, logFile, supportedBgr);
            if (textureFilter == 1) {
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
            } else {
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
            }

            if (textureFileName2[0] != 0) {
                texture2 = LoadTexture(textureFileName2, logFile, supportedBgr);
                if (textureFilter == 1) {
                    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
                    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
                } else {
                    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
                    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
                }
            }
                
            glBindTexture(GL_TEXTURE_2D, texture1);
        }
 #endif
        
        if (depthBits) {
        	glDepthFunc(GL_LEQUAL); if (depthTestLogging) fprintf(logFile, "        glDepthFunc(GL_LEQUAL);\n");
		}

        glMatrixMode(GL_PROJECTION);
        glPushMatrix();
        glOrtho(0.0, windowWidth, 0.0, windowHeight, -1.0, 1.0);
        glMatrixMode(GL_MODELVIEW);
        {
          int sideLength = sqrt(trianglesCount / 2);
          int numIndices = PlaneGetIndicesCount(sideLength + 1, sideLength + 1);
          float *vertices = PlaneGetVertices(sideLength + 1, sideLength + 1);
          int *indices = PlaneGetIndices(sideLength + 1, sideLength + 1);
          int plane, i;
          float angleStep, alpha;
          int numPlanes = scene3_planesCount;

          circleRadius = 20.0f * ((float)windowHeight / 900.0f);          

          if (numPlanes == 0)
              numPlanes = 1;
          angleStep = (2.0f * PI) / (float)numPlanes;
          
          if (blending) {
              glEnable(GL_BLEND);
              glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
          }
              
          if (!displayLists) {
              displayLists = malloc(numPlanes * sizeof(GLuint));
              memset(displayLists, 0, numPlanes * sizeof(GLuint));
              displayListsCount = numPlanes;
          }
              
          alpha = 1.0f / (float)numPlanes;
          for (plane = 0; plane < numPlanes; plane++) {
              glPushMatrix();
              glTranslatef(sinf(param + plane * angleStep) * circleRadius + circleRadius, cosf(param + plane * angleStep) * circleRadius + circleRadius, 0.0f);
              //fprintf(logFile, "circleRadius=%f\n", circleRadius);
              glScalef(((float)windowWidth - 2.0f * circleRadius) / (float)sideLength, ((float)windowHeight - 2.0f * circleRadius) / (float)sideLength, 0.0f);

              if (!displayLists[plane]) {
                  int c;
                  float r,g,b;
                  displayLists[plane] = glGenLists(1);
                  glNewList(displayLists[plane], GL_COMPILE);
                  
                  c = plane % 6;                        // R, G, B, RG, GB, RB
                  r = c == 0 || c == 3 || c == 5 ? 1.0f : (textureFileName1[0] != 0 ? 0.5f : 0.0f);
                  g = c == 1 || c == 3 || c == 4 ? 1.0f : (textureFileName1[0] != 0 ? 0.5f : 0.0f);
                  b = c == 2 || c == 4 || c == 5 ? 1.0f : (textureFileName1[0] != 0 ? 0.5f : 0.0f);
                  
                  if (wireFrame > 0) {
                      glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
                      if (wireFrame == 2){
                          glEnable(GL_LINE_SMOOTH);
                          glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
                          glEnable(GL_BLEND);
                          glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
                      }
                  } else {
                      if (!blending)
                          glDisable(GL_BLEND);
                      glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
                  }
                  
                  if (triangleStrip) {
                      glBegin(GL_TRIANGLE_STRIP);
                          for (i = 0; i < numIndices; i++) {
                              float colorIntensity = (float)i / (float)numIndices;
                              if (!polygonNoColor)
                                  glColor4f(r * colorIntensity, g * colorIntensity, b * colorIntensity, alpha);
                              glVertex2f(vertices[indices[i] * 2], vertices[indices[i] * 2 + 1]);
                          }
                      glEnd();
                  } else {
                      if (textureFileName2[0] != 0 && isMultitexture) {
  #ifndef OGL1
  #ifndef SGI
                          glActiveTextureARB(GL_TEXTURE0_ARB);                        // http://jerome.jouvie.free.fr/opengl-tutorials/Tutorial22.php
                          glEnable(GL_TEXTURE_2D);
                          glBindTexture(GL_TEXTURE_2D, texture1);
                          glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
                          glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
                          glEnable(GL_TEXTURE_GEN_S);
                          glEnable(GL_TEXTURE_GEN_T);
       
                          glActiveTextureARB(GL_TEXTURE1_ARB);
                          glEnable(GL_TEXTURE_2D);
                          glBindTexture(GL_TEXTURE_2D, texture2);
                          glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
                          glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
                          glEnable(GL_TEXTURE_GEN_S);
                          glEnable(GL_TEXTURE_GEN_T);
  #endif
  #endif
                       }
                      
                      glBegin(GL_TRIANGLES);
                      {
                          int row, col;
                          for (row = 0; row < sideLength; row++) {
                              for (col = 0; col < sideLength; col++) {
                                  float colorIntensity = (float)(row * sideLength + (sideLength - col)) / (float)(sideLength * sideLength);

                                  if (polygonNoColor)
                                      glColor4f(1.0f, 1.0f, 1.0f, alpha);


                                  if (!polygonNoColor)
                                      glColor4f(r * colorIntensity, g * colorIntensity, b * colorIntensity, alpha);
  #ifndef OGL1
                                  if (textureFileName2[0] != 0 && isMultitexture) {
  #ifndef SGI
                                      glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.0f, 0.0f);
                                      glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 0.0f, 0.0f);
  #endif
                                  } else {
                                      glTexCoord2f(0.0f, 0.0f);
                                  }
  #endif
                                  glVertex2f((float)col, (float)row);

                                  if (!polygonNoColor)
                                      glColor4f(0.0f, 0.0f, 0.0f, 0.0f);
  #ifndef OGL1
                                  if (textureFileName2[0] != 0 && isMultitexture) {
  #ifndef SGI
                                      glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1.0f, 0.0f);
                                      glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 1.0f, 0.0f);
  #endif
                                  } else {
                                      glTexCoord2f(1.0f, 0.0f);
                                  }
  #endif
                                  glVertex2f((float)(col + 1), (float)row);

                                  if (!polygonNoColor)
                                      glColor4f(r * colorIntensity, g * colorIntensity, b * colorIntensity, alpha);
  #ifndef OGL1
                                  if (textureFileName2[0] != 0 && isMultitexture) {
  #ifndef SGI
                                      glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.0f, 1.0f);
                                      glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 0.0f, 1.0f);
  #endif
                                  } else {
                                      glTexCoord2f(0.0f, 1.0f);
                                  }
  #endif
                                  glVertex2f((float)col, (float)(row + 1));

  #ifndef OGL1
                                  if (textureFileName2[0] != 0 && isMultitexture) {
  #ifndef SGI
                                      glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.0f, 1.0f);
                                      glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 0.0f, 1.0f);
  #endif
                                  } else {
                                      glTexCoord2f(0.0f, 1.0f);
                                  }
  #endif
                                  glVertex2f((float)col, (float)(row + 1));

                                  if (!polygonNoColor)
                                      glColor4f(0.0f, 0.0f, 0.0f, 0.0f);
  #ifndef OGL1
                                  if (textureFileName2[0] != 0 && isMultitexture) {
  #ifndef SGI
                                      glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1.0f, 0.0f);
                                      glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 1.0f, 0.0f);
  #endif
                                  } else {
                                      glTexCoord2f(1.0f, 0.0f);
                                  }
  #endif
                                  glVertex2f((float)(col + 1), (float)row);

                                  if (!polygonNoColor)
                                      glColor4f(r * colorIntensity, g * colorIntensity, b * colorIntensity, alpha);
  #ifndef OGL1
                                  if (textureFileName2[0] != 0 && isMultitexture) {
  #ifndef SGI
                                      glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1.0f, 1.0f);
                                      glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 1.0f, 1.0f);
  #endif
                                  } else {
                                      glTexCoord2f(1.0f, 1.0f);
                                  }
  #endif
                                  glVertex2f((float)(col + 1), (float)(row + 1));
                              }
                          }
                      }
                      glEnd();


                      if (textureFileName2[0] != 0 && !isMultitexture) {
                          glEnable(GL_BLEND);
                          glBlendFunc(GL_ZERO, GL_SRC_COLOR);
  #ifndef OGL1
                          glBindTexture(GL_TEXTURE_2D, texture2);
  #endif                        
                          glBegin(GL_TRIANGLES);
                          {
                              int row, col;
                              for (row = 0; row < sideLength; row++) {
                                  for (col = 0; col < sideLength; col++) {
                                      glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
  #ifndef OGL1    
                                      glTexCoord2f(0.0f, 0.0f);
  #endif
                                      glVertex2f((float)col, (float)row);
  #ifndef OGL1    
                                      glTexCoord2f(1.0f, 0.0f);
  #endif
                                      glVertex2f((float)(col + 1), (float)row);
  #ifndef OGL1    
                                      glTexCoord2f(0.0f, 1.0f);
  #endif
                                      glVertex2f((float)col, (float)(row + 1));
      
  #ifndef OGL1    
                                      glTexCoord2f(0.0f, 1.0f);
  #endif
                                      glVertex2f((float)col, (float)(row + 1));
  #ifndef OGL1    
                                      glTexCoord2f(1.0f, 0.0f);
  #endif
                                      glVertex2f((float)(col + 1), (float)row);
  #ifndef OGL1    
                                      glTexCoord2f(1.0f, 1.0f);
  #endif
                                      glVertex2f((float)(col + 1), (float)(row + 1));
                                  }
                              }
                          }
                          glEnd();
  #ifndef OGL1
                          glBindTexture(GL_TEXTURE_2D, texture1);
  #endif
                          glDisable(GL_BLEND);
                          glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
                      }
                  }
          
                  if (wireFrame < 0) {
                        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
                        if (wireFrame == -2) {
                          glEnable(GL_LINE_SMOOTH);
                          glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
                          glEnable(GL_BLEND);
                          glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
                      }
                      glBegin(GL_TRIANGLE_STRIP);
                          for (i = 0; i < numIndices; i++) {
                              glColor3f(1.0f, 1.0f, 1.0f);
                              glVertex2f(vertices[indices[i] * 2], vertices[indices[i] * 2 + 1]);
                          }
                      glEnd();
                  }
                  
                  glEndList();
              }
              glCallList(displayLists[plane]);
              
              glPopMatrix();
          }
          
          if (checkDephtTestFill) {
                  glBegin(GL_QUADS);
                          glColor3f(0.0f, 0.0f, 1.0f);
                          glVertex3f((float)windowWidth * 0.25f, (float)windowHeight * 0.35f, -0.5f);
                          glColor3f(1.0f, 1.0f, 1.0f);
                          glVertex3f((float)windowWidth * 0.75f, (float)windowHeight * 0.15f, 0.5f);
                          glVertex3f((float)windowWidth * 0.75f, (float)windowHeight * 0.85f, 0.5f);
                          glColor3f(0.0f, 0.0f, 1.0f);
                          glVertex3f((float)windowWidth * 0.25f, (float)windowHeight * 0.65f, -0.5f);
                  glEnd();
                  }
          if (checkDephtTestWireframe) {
                          glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
                          if (checkDephtTestWireframe == 2) {
                                  glEnable(GL_LINE_SMOOTH);
                                  glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
                                  glEnable(GL_BLEND);
                                  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
                          }

                          if (depthBits) {
                                  glDisable(GL_DEPTH_TEST); if (depthTestLogging) fprintf(logFile, "    glDisable(GL_DEPTH_TEST);\n");
                          }
                  glBegin(GL_QUADS);
                          glColor3f(1.0f, 1.0f, 1.0f);
                          glVertex3f((float)windowWidth * 0.25f, (float)windowHeight * 0.35f, -0.5f);
                          glVertex3f((float)windowWidth * 0.75f, (float)windowHeight * 0.15f, 0.5f);
                          glVertex3f((float)windowWidth * 0.75f, (float)windowHeight * 0.85f, 0.5f);
                          glVertex3f((float)windowWidth * 0.25f, (float)windowHeight * 0.65f, -0.5f);
                  glEnd();
                          if (depthBits) {
                                  glEnable(GL_DEPTH_TEST); if (depthTestLogging) fprintf(logFile, "    glEnable(GL_DEPTH_TEST);\n");
                          }
                  
                          glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
                          if (checkDephtTestWireframe == 2)
                                  glDisable(GL_BLEND);
                  }

          glMatrixMode(GL_PROJECTION);
          glPopMatrix();
          glMatrixMode(GL_MODELVIEW);
       }
    }
}

void DeInit() {
    int i;
    if (displayLists) {
        for (i = 0; i < displayListsCount; i++)
            glDeleteLists(displayLists[i], 1);
        free(displayLists);
    }
    
    PlaneFreeData();
#ifndef OGL1
    if (textureFileName1[0] != 0 && texture1 == 0)
        glDeleteTextures(1, &texture1);
    if (textureFileName2[0] != 0 && texture2 == 0)
        glDeleteTextures(1, &texture2);
#endif

#ifndef X11    
    wglMakeCurrent(NULL, NULL);
    wglDeleteContext(hRC);
    ReleaseDC(hWnd, hDC);

    DestroyWindow(hWnd);
#endif

    if (csvFileName[0] != 0) {
        fprintf(csvFile, ",");
        if (fps != -1.0f)
            fprintf(csvFile, "%f", fps);
        
        fprintf(csvFile, ",");
        if (scene == 3 && fps != -1.0f)
            fprintf(csvFile, "%.0f", ((float)windowWidth - 2.0f * circleRadius) * ((float)windowHeight - 2.0f * circleRadius) * scene3_planesCount * fps);
        else if (scene == 2 && fps != -1.0f)
            fprintf(csvFile, "%.0f", (float)realTrianglesCount * (float)scene2_trianglesPixels * (float)fps);

        fprintf(csvFile, ",");
        if (scene == 3 && fps != -1.0f)
            fprintf(csvFile, "%.0f", realTrianglesCount * scene3_planesCount * fps);
        else if (scene == 2 && fps != -1.0f)
            fprintf(csvFile, "%.0f", realTrianglesCount * fps);

        fprintf(csvFile, ",%i,%i,%i,%i,%i," /*"%i,%i,"*/ "%i,%i,%i,%i,"
            "%i,%i,%i,%i,%i,%i,%s,%s,%i,%i,%i,"
            "%i", 
            windowWidth, windowHeight, colorsBits, depthBits, windowFrame, /*timeDemo, targetFps,*/ clearBuffers, scene, trianglesCount, realTrianglesCount,
            scene2_trianglesPixels, scene3_planesCount, wireFrame, blending, runSeconds, triangleStrip, textureFileName1, textureFileName2, polygonNoColor, textureFilter, isMultitexture,
            disableMultitexture);
        
        fprintf(csvFile, "\n");
        fclose(csvFile);
    }

    fprintf(logFile, "FPS %.2f\n", fps);    
    fclose(logFile);
}

#ifndef X11
LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam) {
    switch (message)
    {
    case WM_CREATE:
        return 0;
    case WM_CLOSE:
        PostQuitMessage(0);
        return 0;
    case WM_DESTROY:
        return 0;
    case WM_KEYDOWN:
        switch (wParam)
        {
        case VK_ESCAPE:
            PostQuitMessage(0);
            return 0;
        }
        return 0;
    default:
        return DefWindowProc(hWnd, message, wParam, lParam);
    }
}
#else
int process_input(Display *dpy) {
    XEvent event;
    char buf[31];
    KeySym keysym;

    if (!XPending(dpy))
        return 1;

    XNextEvent(dpy, &event);
    switch(event.type) {
    case Expose:
        break;
    case ConfigureNotify:
        glViewport(0, 0, event.xconfigure.width, event.xconfigure.height);
        break;
    case KeyPress:
        (void) XLookupString(&event.xkey, buf, sizeof(buf), &keysym, NULL);
        switch (keysym) {
        case XK_Escape: case 'Q': case 'q':
            /*exit(EXIT_SUCCESS)*/ return 0;
        default:
            break;
        }
    default:
        break;
    }
    return 1;
}
#endif
