// http://www.opengl-tutorial.org/beginners-tutorials/tutorial-5-a-textured-cube/#loading-bmp-images-yourself

#include <stdlib.h>
#include <stdio.h>
#include <GL/gl.h>
#include "glext.h"

GLuint LoadTexture(const char *path, FILE *logFile, int supportedBrg) {
    // Data read from the header of the BMP file
    unsigned char header[54]; // Each BMP file begins by a 54-bytes header
    unsigned int dataPos;     // Position in the file where the actual data begins
    unsigned int width, height;
    unsigned int imageSize;   // = width*height*3
    // Actual RGB data
    unsigned char *data;
    GLuint textureID;
    
    // Open the file
    FILE *file = fopen(path, "rb");
    if (!file) {
        fprintf(logFile, "Image could not be opened\n"); 
        return 0;
    }

    if ( fread(header, 1, 54, file)!=54 ){ // If not 54 bytes read : problem
        fprintf(logFile, "Not a correct BMP file\n");
        return 0;
    }

    if ( header[0]!='B' || header[1]!='M' ){
        fprintf(logFile, "Not a correct BMP file\n");
        return 0;
    }
    
    // Read ints from the byte array
    //dataPos    = *(int*)&(header[0x0A]);
    //imageSize  = *(int*)&(header[0x22]);

    //width      = *(int*)&header[18];            -- crashes on old SGI O2 (1996)
    width      = header[18] | (header[19] << 8) | (header[20] << 16) | (header[21] << 24);

    //height     = *(int*)&header[22];            -- crashes on old SGI O2 (1996)
    height      = header[22] | (header[23] << 8) | (header[24] << 16) | (header[25] << 24);

    imageSize = width * height * 3;
    
    // Some BMP files are misformatted, guess missing information
    //if (imageSize==0)    imageSize=width*height*3; // 3 : one byte for each Red, Green and Blue component
    //if (dataPos==0)      dataPos=54; // The BMP header is done that way
    
    // Create a buffer
    data = malloc(sizeof(unsigned char) * imageSize);
    
    // Read the actual data from the file into the buffer
    fread(data,1,imageSize,file);
    
    //Everything is in memory now, the file can be closed
    fclose(file);
    
    // Create one OpenGL texture
    glGenTextures(1, &textureID);
    
    // "Bind" the newly created texture : all future texture functions will modify this texture
    glBindTexture(GL_TEXTURE_2D, textureID);            // GL_EXT_texture_object nebo OpenGL 1.1 -- https://www.khronos.org/registry/OpenGL/extensions/EXT/EXT_texture_object.txt
    
    // Give the image to OpenGL
    if (supportedBrg) {
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width, height, 0, GL_BGR, GL_UNSIGNED_BYTE, data);
    } else {
        unsigned char *data_rgb = malloc(width * height * 3), *p = data, *p_rgb = data_rgb;
        int i;
        for (i = 0; i < width * height * 3; i += 3) {
            *(p_rgb) = *(p + 2);
            *(p_rgb + 1) = *(p + 1);
            *(p_rgb + 2) = *(p);
            p += 3, p_rgb += 3;
        }
        
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, data_rgb);
        
        free(data_rgb);
    }
    free(data);
    
    return textureID;
}
